"""Implements methods for solving scalar valued functions.

"""
from typing import Any, Callable, Dict, Optional, Tuple, Union

import numpy as np
from scipy.optimize import NonlinearConstraint, differential_evolution, minimize

from desdeo_tools.scalarization.Scalarizer import DiscreteScalarizer, Scalarizer


class ScalarSolverException(Exception):
    pass


class ScalarMethod:
    """A class the define and implement methods for minimizing scalar valued functions.
    """

    def __init__(self, method: Callable, method_args=None, use_scipy: Optional[bool] = False):
        """
        Args:
            method (Callable): A callable minimizer function which expects a
            callable scalar valued function to be minimized. The function should
            accept as its first argument a two dimensional numpy array and should
            return a dictionary with at least the keys: "x" the found optimal solution,
            "success" boolean indicating if the minimization was successfull,
            "message" a string of additional info.
            method_args (Dict, optional): Any other keyword arguments to be supplied
            to the method. Defaults to None.
            use_scipy (Optional[bool]): Whether to use scipy's NonLinearConstraint to
            handle the constraints.
        """
        self._method = method
        self._method_args = method_args
        self._use_scipy = use_scipy

    def __call__(self, obj_fun: Callable, x0: np.ndarray, bounds: np.ndarray, constraint_evaluator: Callable) -> Dict:
        """Minimizes a scalar valued function.
        
        Args:
            obj_fun (Callable): A callable scalar valued function that
            accepts a two dimensional numpy array as its first arguments.
            x0 (np.ndarray): An initial guess.
            bounds (np.ndarray): The upper and lower bounds for each variable
            accepted by obj_fun. Expects a 2D numpy array with each row
            representing the lower and upper bounds of a variable. The first column
            should contain the lower bounds and the last column the upper bounds.
            Use np.inf to indicate no bound.
            constraint_evaluator (Callable): Should accepts exactly the
            same arguments as obj_fun. Returns a scalar value for each constraint
            present. This scalar value should be positive if a constraint holds, and negative
            otherwise.
        
        Returns:
            Dict: A dictionary with at least the following entries: 'x' indicating the optimal
            variables found, 'fun' the optimal value of the optimized function, and 'success' a boolean
            indicating whether the optimization was conducted successfully.
        """
        if self._method_args is not None:
            res = self._method(obj_fun, x0, bounds=bounds, constraints=constraint_evaluator, **self._method_args)
        else:
            res = self._method(obj_fun, x0, bounds=bounds, constraints=constraint_evaluator)

        return res


class ScalarMinimizer:
    """Implements a class for minimizing scalar valued functions with bounds set for the
    variables, and constraints.
    """

    def __init__(
        self,
        scalarizer: Scalarizer,
        bounds: np.ndarray,
        constraint_evaluator: Callable = None,
        method: Optional[Union[ScalarMethod, str]] = None,
    ):
        """ 
        Args:
            scalarizer (Scalarizer): A Scalarizer to be minimized.
            bounds (np.ndarray): The bounds of the independent variables the
            scalarizer is called with.
            constraint_evaluator (Callable, optional): A Callable which
            representing a vector valued constraint function. The array the constraint
            function returns should be two dimensional with each row corresponding to the
            constraint function values when evaluated. A value of less than zero is
            understood as a non valid constraint. Defaults to None.
            method (Optional[Union[Callable, str]], optional): The optimization method the scalarizer
            should be minimized with. It should accepts as keyword the arguments 'bounds' and 
            'constraints' which will be used to pass it the bounds and constraint_evaluator.
            If none is supplied, uses the minimizer implemented in SciPy. Otherwise a str can be given
            to use one of the preset solvers available. Use the method 'get_presets' to get a list
            of available preset solvers.
            Defaults to None.
        """
        self.presets = ["scipy_minimize", "scipy_de"]

        self._scalarizer = scalarizer
        self._bounds = bounds
        self._constraint_evaluator = constraint_evaluator
        if (method is None) or (method == "scipy_minimize"):
            # scipy minimize
            self._use_scipy = True
            # Assuming the gradient reqruies evaluation of the
            # scalarized function with out of bounds variable values.
            self._bounds[:, 0] += 1e-6
            self._bounds[:, 1] -= 1e-6
            self._method = ScalarMethod(minimize)

        elif method == "scipy_de":
            # Scipy differential evolution
            self._use_scipy = True
            # Assuming the gradient reqruies evaluation of the
            # scalarized function with out of bounds variable values.
            # only relevant if the 'polish' option is set in scipy's DE
            self._bounds[:, 0] += 1e-6
            self._bounds[:, 1] -= 1e-6
            scipy_de_method = ScalarMethod(
                lambda x, _, **y: differential_evolution(x, **y), method_args={"polish": True}
            )
            self._method = scipy_de_method

        else:
            self._use_scipy = method._use_scipy
            self._method = method

            if self._use_scipy:
                # Assuming the gradient reqruies evaluation of the
                # scalarized function with out of bounds variable values.
                # only relevant if the 'polish' option is set in scipy's DE
                self._bounds[:, 0] += 1e-6
                self._bounds[:, 1] -= 1e-6

    def get_presets(self):
        """Return the list of preset minimizers available.

        """
        return self.get_presets

    def minimize(self, x0: np.ndarray) -> Dict:
        """Minimizes the scalarizer given an initial guess x0.
        
        Args:
            x0 (np.ndarray): A numpy array containing an initial guess of variable values.

        Returns:
            Dict: A dictionary with at least the following entries: 'x' indicating the optimal
            variables found, 'fun' the optimal value of the optimized function, and 'success' a boolean
            indicating whether the optimizaton was conducted successfully.
        """
        if self._use_scipy:
            # create wrapper for the constraints to be used with scipy's minimize routine.
            # assuming that all constraints hold when they return a positive value.
            if self._constraint_evaluator is not None:
                scipy_cons = NonlinearConstraint(self._constraint_evaluator, 0, np.inf)
            else:
                scipy_cons = ()

            res = self._method(self._scalarizer, x0, bounds=self._bounds, constraint_evaluator=scipy_cons)

        else:
            res = self._method(
                self._scalarizer, x0, bounds=self._bounds, constraint_evaluator=self._constraint_evaluator
            )

        return res


class DiscreteMinimizer:
    """Implements a class for finding the minimum value of a discrete of scalarized vectors.
    
    """

    def __init__(
        self,
        discrete_scalarizer: DiscreteScalarizer,
        constraint_evaluator: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ):
        """
        Args:
            discrete_scalarizer (DiscreteScalarizer): A discrete scalarizer
            which takes as its arguments an array of vectors and returns a
            scalar value for each vector.
            constraint_evaluator (Optional[Callable[[np.ndarray],
            np.ndarray]], optional): An evaluator which returns True if a
            given vector(s) adheres to given constraints, and False
            otherwise. Defaults to None.
        """
        self._scalarizer = discrete_scalarizer
        self._constraint_evaluator = constraint_evaluator

    def minimize(self, vectors: np.ndarray) -> int:
        """Find the index of the element in vectors which minimizes the
        scalar value returned by the scalarizer. If multiple minimum values
        are found, returns the index of the first occurrence.
        
        Args:
            vectors (np.ndarray): The vectors for which the minimum scalar
            value should be computed for.
        
        Raises:
            ScalarSolverException: None of the given vectors adhere to the
            given constraints.
        
        Returns:
            int: The index of the vector in vectors which minimizes the value
            computed with the given scalarizer.
        """
        if self._constraint_evaluator is None:
            return np.nanargmin(self._scalarizer(vectors))
        else:
            bad_con_mask = ~self._constraint_evaluator(vectors)
            if np.all(bad_con_mask):
                raise ScalarSolverException("None of the supplied vectors adhere to the given " "constraint function.")
            tmp = np.copy(vectors)
            tmp[bad_con_mask] = np.nan
            return np.nanargmin(self._scalarizer(tmp))


if __name__ == "__main__":
    from desdeo_tools.scalarization.ASF import PointMethodASF

    ideal = np.array([0, 0, 0, 0])
    nadir = np.array([1, 1, 1, 1])

    asf = PointMethodASF(nadir, ideal)
    dscalarizer = DiscreteScalarizer(asf, {"reference_point": None})
    dminimizer = DiscreteMinimizer(dscalarizer)

    non_dominated_points = np.array(
        [[0.2, 0.4, 0.6, 0.8], [0.4, 0.2, 0.6, 0.8], [0.6, 0.4, 0.2, 0.8], [0.4, 0.8, 0.6, 0.2]]
    )

    z = np.array([0.4, 0.2, 0.6, 0.8])

    dscalarizer._scalarizer_args = {"reference_point": z}

    print(asf(non_dominated_points, reference_point=z))

    res = dminimizer.minimize(non_dominated_points)
    print(non_dominated_points[res])



import abc
from abc import abstractmethod
from os import path
from typing import List, Union

import numpy as np


class MOEADSFError(Exception):
    """Raised when an error related to the MOEADSF classes is encountered.

    """


class MOEADSFBase(abc.ABC):
    """A base class for representing scalarizing functions for the MOEA/D algorithm.
    Instances of the implementations of this class should work as
    function.

    """

    @abstractmethod
    def __call__(
        self,
        objective_vector: np.ndarray,
        reference_vector: np.ndarray,
        ideal_vector: np.ndarray,
        nadir_vector: np.ndarray,
    ) -> Union[float, np.ndarray]:
        """Evaluate the SF.

        Args:
            objective_vector (np.ndarray): The objective vector to calculate
            the values.
            reference_vector (np.ndarray): The reference vector to calculate the
            values.
            ideal_vector (np.ndarray): The ideal objective vector.
            nadir_vector (np.ndarray): The nadir objective vector.
        Returns:
            Union[float, np.ndarray]: Either a single SF value or a vector of
            values if objective is a 2D array.
        """
        pass


class Tchebycheff(MOEADSFBase):
    """Implements the Tchebycheff scalarizing function.
    """

    def __call__(
        self,
        objective_vector: np.ndarray,
        reference_vector: np.ndarray,
        ideal_vector: np.ndarray,
    ) -> Union[float, np.ndarray]:
        """Evaluate the Tchebycheff scalarizing function for minimization problems.

        Args:
            objective_vector (np.ndarray): A vector representing a solution in
            the objective space.
            reference_vector (np.ndarray): A reference vector representing the direction
            ideal_vector (np.ndarray): The ideal objective vector 
        Raises:
            MOEADSFError: The dimensions of the objective vector and reference_vector don't match.

        Note:
            The shaped of objective_vector and reference_vector must match.

        """
        if not objective_vector.shape == reference_vector.shape:
            msg = (
                "The dimensions of the objective vector {} and "
                "reference_vector {} do not match."
            ).format(objective_vector.shape, reference_vector.shape)
            raise MOEADSFError(msg)

        feval = np.abs(objective_vector - ideal_vector) * reference_vector
        max_fun = np.max(feval)
        return max_fun


class WeightedSum(MOEADSFBase):
    """Implements the Weighted sum scalarization function
    """

    def __call__(
        self, objective_vector: np.ndarray, reference_vector: np.ndarray
    ) -> Union[float, np.ndarray]:
        """Evaluate the WeightedSum scalarizing function.

        Args:
            objective_vector (np.ndarray): A vector representing a solution in
            the objective space.
            reference_vector (np.ndarray): A reference vector representing the direction
        Raises:
            MOEADSFError: The dimensions of the objective vector and reference_vector don't match.

        Note:
            The shaped of objective_vector and reference_vector must match. A reference point is not needed.

        """
        if not objective_vector.shape == reference_vector.shape:
            msg = (
                "The dimensions of the objective vector {} and "
                "reference_vector {} do not match."
            ).format(objective_vector.shape, reference_vector.shape)
            raise MOEADSFError(msg)
        feval = np.sum(objective_vector * reference_vector)
        return feval


class PBI(MOEADSFBase):
    """Implements the PBI scalarization function
    Args:
        theta(float): A penalty parameter used by the function

    Attributes:
        theta (float): A penalty parameter used by the function

    .. Q. Zhang and H. Li, "MOEA/D: A Multiobjective Evolutionary Algorithm Based on Decomposition," 
       in IEEE Transactions on Evolutionary Computation, vol. 11, no. 6, pp. 712-731, Dec. 2007, doi: 10.1109/TEVC.2007.892759.
    """

    def __init__(self, theta: float = 5):
        self.theta = theta

    def __call__(
        self,
        objective_vector: np.ndarray,
        reference_vector: np.ndarray,
        ideal_vector: np.ndarray,
    ) -> Union[float, np.ndarray]:
        """Evaluate the PBI scalarizing function for minimization problems.

        Args:
            objective_vector (np.ndarray): A vector representing a solution in
            the objective space.
            reference_vector (np.ndarray): A reference vector representing the direction
            ideal_vector (np.ndarray): The ideal objecive vector
        Raises:
            MOEADSFError: The dimensions of the objective vector and reference_vector don't match.

        Note:
            The shaped of objective_vector and reference_vector must match. The reference point is not needed.

        """
        if not objective_vector.shape == reference_vector.shape:
            msg = (
                "The dimensions of the objective vector {} and "
                "reference_vector {} do not match."
            ).format(objective_vector.shape, reference_vector.shape)
            raise MOEADSFError(msg)

        norm_weights = np.linalg.norm(reference_vector)
        weights = np.true_divide(reference_vector, norm_weights)
        fx_a = objective_vector - ideal_vector
        d1 = np.fabs(np.inner(fx_a, weights))

        fx_b = objective_vector - (ideal_vector + d1 * weights)
        d2 = np.linalg.norm(fx_b)

        fvalue = d1 + self.theta * d2
        return fvalue




"""This module contains classes implementing different interactions to be
used to communicate between different optimization algorithms and users.

"""

__all__ = [
    "PrintRequest",
    "SimplePlotRequest",
    "ReferencePointPreference",
    "PreferredSolutionPreference",
    "NonPreferredSolutionPreference",
    "BoundPreference",
    "validate_ref_point_data_type",
    "validate_ref_point_dimensions",
    "validate_ref_point_with_ideal",
    "validate_ref_point_with_ideal_and_nadir",
    "validate_with_ref_point_nadir",
    "validate_specified_solutions",
    "validate_bounds",
]

from desdeo_tools.interaction.request import (
    PrintRequest,
    ReferencePointPreference,
    SimplePlotRequest,
    PreferredSolutionPreference,
    NonPreferredSolutionPreference,
    BoundPreference,
)
from desdeo_tools.interaction.validators import (
    validate_ref_point_data_type,
    validate_ref_point_dimensions,
    validate_ref_point_with_ideal,
    validate_ref_point_with_ideal_and_nadir,
    validate_with_ref_point_nadir,
    validate_specified_solutions,
    validate_bounds,
)



from typing import Any, Callable, Dict, Optional

import numpy as np


class Scalarizer:
    """Implements a class for scalarizing vector valued functions with a
    given scalarization function.
    """

    def __init__(
        self, evaluator: Callable, scalarizer: Callable, evaluator_args: Dict = None, scalarizer_args: Dict = None
    ):
        """
        Args:
            evaluator (Callable): A Callable object returning a numpy array.
            scalarizer (Callable): A function which should accepts as its
            arguments the output of evaluator and return a single value.
            evaluator_args (Any, optional): Optional arguments to be passed to
            evaluator. Defaults to None.
            scalarizer_args (Any, optional): Optional arguments to be passed to
            scalarizer. Defaults to None.
        """
        self._evaluator = evaluator
        self._scalarizer = scalarizer
        self._evaluator_args = evaluator_args
        self._scalarizer_args = scalarizer_args

    def evaluate(self, xs: np.ndarray) -> np.ndarray:
        """Evaluates the scalarized function with the given arguments and
        returns a scalar value for each vector of variables given in a numpy
        array.
        
        Args:
            xs (np.ndarray): A 2D numpy array containing vectors of variables
            on each of its rows.
        
        Returns:
            np.ndarray: A 1D numpy array with the values returned by the
            scalarizer for each row in xs.
        """
        if self._evaluator_args is not None:
            res_eval = self._evaluator(xs, **self._evaluator_args)
        else:
            res_eval = self._evaluator(xs)

        if self._scalarizer_args is not None:
            res_scal = self._scalarizer(res_eval, **self._scalarizer_args)
        else:
            res_scal = self._scalarizer(res_eval)

        return res_scal

    def __call__(self, xs: np.ndarray) -> np.ndarray:
        """Wrapper to the evaluate method.
        """
        return self.evaluate(xs)


class DiscreteScalarizer:
    """Implements a class to scalarize discrete vectors given a scalarizing function.

    """

    def __init__(self, scalarizer: Callable, scalarizer_args: Dict = None):
        self._scalarizer = scalarizer
        self._scalarizer_args = scalarizer_args

    def evaluate(self, vectors: np.ndarray) -> np.ndarray:
        # guarantee two dimensions, makes sure works with single vectors as well.
        vectors = np.atleast_2d(vectors)

        if self._scalarizer_args is not None:
            res_scal = self._scalarizer(vectors, **self._scalarizer_args)
        else:
            res_scal = self._scalarizer(vectors)

        return res_scal

    def __call__(self, vectors: np.ndarray):
        return self.evaluate(vectors)


if __name__ == "__main__":
    vectors = np.array([[1, 1, 1], [2, 2, 2], [4, 5, 6.0]])
    vector = np.array([1, 2, 3])
    dscalarizer = DiscreteScalarizer(lambda x, a=1: a * np.sum(x, axis=1), scalarizer_args={"a": 2})
    res = dscalarizer(vectors)
    res_1d = dscalarizer(vector)
    print(res)
    print(res_1d)
    pass



import numba

import numpy as np

from typing import List


@numba.njit()
def dominates(x: np.ndarray, y: np.ndarray) -> bool:
    """Returns true if x dominates y.

    Args:
        x (np.ndarray): First solution. Should be a 1-D array of numerics.
        y (np.ndarray): Second solution. Should be the same shape as x.

    Returns:
        bool: True if x dominates y, false otherwise.
    """
    dom = False
    for i in range(len(x)):
        if x[i] > y[i]:
            return False
        elif x[i] < y[i]:
            dom = True
    return dom


@numba.njit()
def non_dominated(data: np.ndarray) -> np.ndarray:
    """Finds the non-dominated front from a population of solutions.

    Args:
        data (np.ndarray): 2-D array of solutions, with each row being a single solution.

    Returns:
        np.ndarray: Boolean array of same length as number of solutions (rows). The value is
            true if corresponding solution is non-dominated. False otherwise
    """

    num_solutions = len(data)
    index = np.zeros(num_solutions, dtype=np.bool_)
    index[0] = True
    for i in range(1, num_solutions):
        index[i] = True
        for j in range(i):
            if not index[j]:
                continue
            if dominates(data[i], data[j]):
                index[j] = False
            elif dominates(data[j], data[i]):
                index[i] = False
                break
    return index


@numba.njit()
def fast_non_dominated_sort(data: np.ndarray) -> np.ndarray:
    """Conduct fast non-dominated sorting on a population of solutions.

    Args:
        data (np.ndarray): 2-D array of solutions, with each row being a single solution.

    Returns:
        np.ndarray: n x f boolean array. n is the number of solutions, f is the number of fronts.
            The value of an array element is true if the corresponding solution id (column) belongs in
            the corresponding front (row).
    """

    num_solutions = len(data)
    indices = np.arange(num_solutions)
    taken = np.zeros(num_solutions, dtype=np.bool_)
    fronts = np.zeros((num_solutions, num_solutions), dtype=np.bool_)

    for i in indices:
        current_front = non_dominated(data[~taken])

        current_front_all = np.zeros(num_solutions, dtype=np.bool_)
        current_front_all[~taken] = current_front
        fronts[i] = current_front_all

        taken = taken + fronts[i]
        if not fronts[i].any():
            break
    return fronts[:i]


def fast_non_dominated_sort_indices(data: np.ndarray) -> List[np.ndarray]:
    """Conduct fast non-dominated sorting on a population of solutions.

    This function returns identical results as `fast_non_dominated_sort`, but in a different format.
    This function returns an array of solution indices for each front, packed in a list.

    Args:
        data (np.ndarray): 2-D array of solutions, with each row being a single solution.

    Returns:
        List[np.ndarray]: A list with f elements where f is the number of fronts in the data,
            arranged in ascending order. Each element is a numpy array of the indices of solutions
            belonging to the corresponding front.
    """
    fronts = fast_non_dominated_sort(data)
    return [np.where(fronts[i]) for i in range(len(fronts))]



from typing import Callable, List, Union

import pandas as pd

from desdeo_tools.interaction.validators import (
    validate_ref_point_with_ideal_and_nadir,
    validate_specified_solutions,
    validate_bounds,
)
from desdeo_tools.utilities.frozen import FrozenClass


class RequestError(Exception):
    """Raised when an error related to the Request class is encountered.
    """


class BaseRequest(FrozenClass):
    """The base class for all Request classes. Request classes are to be used
    to handle interaction between the user and the methods, as well as within
    various methods. This class is frozen, so no variables other than that
    already defined in current __init__ can be defined in derived classes.
    """

    def __init__(
        self,
        request_type: str,
        interaction_priority: str,
        content=None,
        request_id: int = None,
    ):
        """Initialize a BaseRequest class. This method contains a lot of
        boilerplate.

        Args:
            request_type (str): The type of request. Currently, one of ["print",
            "simple_plot", "reference_point_preference",
            "classification_preference"].
            interaction_priority (str): The priority of preference, as decided
            by the method. One of ["no_interaction", "not_required",
            "recommended", "required"], with trivial meanings.
            content ([type], optional): The data relevant to the request packet.
            For example, if the request type is print, content may contain
            strings to be printed. Typically a dict. Defaults to None.
            request_id (int, optional): A unique identifier. Defaults to None.

        Raises:
            RequestError: If request type is not recognized
            RequestError: If request priority is not recognized
            RequestError: If request id is not an integer
        """
        acceptable_types = [
            "print",
            "simple_plot",
            "reference_point_preference",
            "classification_preference",
            "preferred_solution_preference",
            "non_preferred_solution_preference",
            "bound_preference",
        ]
        priority_types = ["no_interaction", "not_required", "recommended", "required"]
        if request_type not in acceptable_types:
            msg = f"Request type should be one of {acceptable_types}"
            raise RequestError(msg)
        if interaction_priority not in priority_types:
            msg = f"Request priority should be one of {priority_types}"
            raise RequestError(msg)
        if not isinstance(request_id, (int, type(None))):
            msg = "Request id should be int or None"
            raise RequestError(msg)
        # Attributes
        self._request_type: str = request_type
        self._interaction_priority: str = interaction_priority  # Example: one of:
        self._request_id: int = request_id  # Some random number as id
        self._content = content
        self._response = None
        #  Freezing this class
        self._freeze()

    @property
    def request_type(self):
        return self._request_type

    @property
    def interaction_priority(self):
        return self._interaction_priority

    @property
    def request_id(self):
        return self._request_id

    @property
    def content(self):
        return self._content

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self):
        # Code to validate the response
        return


class PrintRequest(BaseRequest):
    """Methods can use this request class to send out textual information to be
    displayed to the decision maker. This could be a single message in the form
    of a string, or multiple messages in a list of strings. The method of
    displaying these messages is left to the UI.
    """

    def __init__(self, message: Union[str, List[str]], request_id: int = None):
        """Initialise the PrintRequest.

        Args:
            message (Union[str, List[str]]): A single message (str) or a list of
            messages to be displayed to the decision maker
            request_id (int, optional): A unique identifier for this request.
            Defaults to None.

        Raises:
            RequestError: If message is not a str or a list
            RequestError: If message is a list but one or more elements are not
            str.
        """
        if not isinstance(message, str):
            if not isinstance(message, list):
                msg = (
                    f"Message/s to be printed should be string or list of strings"
                    f"Message provided is of type: {type(message)}"
                )
                raise RequestError(msg)
            elif not all(isinstance(x, str) for x in message):
                msg = (
                    "Message/s to be printed should be string or list of strings"
                    "Some elements of the list are not strings"
                )
                raise RequestError(msg)
        super().__init__(
            request_type="print",
            interaction_priority="no_interaction",
            content=message,
            request_id=request_id,
        )


class SimplePlotRequest(BaseRequest):
    """Methods can use this request class to send out some data to be shown to
    the decision maker (typically in the form of a plot). This data is usually a
    set of solutions, stored in the content variable of this class. The manner
    of visualization is left to the UI.

    content is a dict that contains the following keys:
    "data" (pandas.DataFrame): The data to be plotted.
    "dimensional_data" (pandas.Dataframe): The data contained in this key can be
    used to scale the data to be plotted.
    "chart_title" (str): A recommended title for the visualization.
    "message" (Union[str, List[str]]): A message or list of messages to be
    displayed to the decision maker.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        message: Union[str, List[str]],
        dimensions_data: pd.DataFrame = None,
        chart_title: str = None,
        request_id=None,
    ):
        """Initialize the request packet

        Args:
            data (pd.DataFrame): The data to be plotted.
            message (Union[str, List[str]]): A message or list of messages to be
            displayed to the decision maker.
            dimensions_data (pd.DataFrame, optional): Data used to used to scale
            the data to be plotted. Defaults to None.
            chart_title (str, optional): A recommended title for the
            visualization. Defaults to None.
            request_id ([type], optional): A unique identifier. Defaults to None.

        Raises:
            RequestError: data is not a pandas DataFrame.
            RequestError: dimensions_data is not a pandas DataFrame or None.
            RequestError: A mismatch in the column names of data and
            dimensions_data.
            RequestError: If dimensions_data DataFrame contains indices other
            that "minimize", "ideal", or "nadir".
            RequestError: If chart_title is not str or None.
            RequestError: If message is not a str or a list.
            RequestError: If message is a list but one or more elements are not
            str.
        """
        acceptable_dimensions_data_indices = [
            "minimize",  # 1 if minimized, -1 if maximized
            "ideal",
            "nadir",
        ]
        if not isinstance(data, pd.DataFrame):
            msg = (
                f"Provided data to be plotted should be in a pandas dataframe, with"
                f"columns names being the same as objective names.\n"
                f"Provided data is of type: {type(data)}"
            )
            raise RequestError(msg)
        if not isinstance(dimensions_data, (pd.DataFrame, type(None))):
            msg = (
                f"Dimensional data should be in a pandas dataframe.\n"
                f"Provided data is of type: {type(dimensions_data)}"
            )
            raise RequestError(msg)
        if not all(data.columns == dimensions_data.columns):
            msg = (
                f"Mismatch in column names of data and dimensions_data.\n"
                f"Column names in data: {data.columns}"
                f"Column names in dimensions_data: {dimensions_data.columns}"
            )
            raise RequestError(msg)
        rouge_indices = [
            index
            for index in dimensions_data.index
            if index not in acceptable_dimensions_data_indices
        ]
        if rouge_indices:
            msg = (
                f"dimensions_data should only contain the following indices:\n"
                f"{acceptable_dimensions_data_indices}\n"
                f"The dataframe provided contains the following unsupported indices:\n"
                f"{rouge_indices}"
            )
            raise RequestError(msg)
        if not isinstance(chart_title, (str, type(None))):
            msg = (
                f"Chart title should be a string. Provided chart type is:"
                f"{type(chart_title)}"
            )
            raise RequestError(msg)
        if not isinstance(message, str):
            if not isinstance(message, list):
                msg = (
                    f"Message/s to be printed should be string or list of strings"
                    f"Message provided is of type: {type(message)}"
                )
                raise RequestError(msg)
            elif not all(isinstance(x, str) for x in message):
                msg = (
                    "Message/s to be printed should be string or list of strings"
                    "Some elements of the list are not strings"
                )
                raise RequestError(msg)
        content = {
            "data": data,
            "dimensions_data": dimensions_data,
            "chart_title": chart_title,
            "message": message,
        }
        super().__init__(
            request_type="simple_plot",
            interaction_priority="no_interaction",
            content=content,
            request_id=request_id,
        )


class ReferencePointPreference(BaseRequest):
    """Methods can use this request class to ask the DM to provide their
    preferences in the form of a reference point. This reference point is
    validated according to the needs of the method that initializes this class
    object, before the reference point can be accepted in the response variable.
    """

    def __init__(
        self,
        dimensions_data: pd.DataFrame,
        message: str = None,
        interaction_priority: str = "required",
        preference_validator: Callable = None,
        request_id: int = None,
    ):
        """Initialize the request class.

        Args:
            dimensions_data (pd.DataFrame): Minimal data that should be shown to
            the decision maker. If a lot of data needs to be shown (i.e., with a
            visualization), use SimplePlotRequest or related classes for that
            purpose, and this class for the interaction with the decision maker.
            message (str, optional): Message to be displayed to a decision
            maker. Defaults to None.
            interaction_priority (str, optional): The importance of the
            interaction as decided by the method. If equal to "required", the
            method will not continue without a DM preference. If equal to
            "recommended", the interaction is recommended, but not required for
            the continuation of the method. The case "not_required" is similar
            to "recommended". Defaults to "required".
            preference_validator (Callable, optional): A callable function that
            tests whether a reference point provided by the DM is valid or not.
            Defaults to None.
            request_id (int, optional): A unique identifier. Defaults to None.

        Raises:
            RequestError: dimensions_data is not a pandas DataFrame.
            RequestError: If dimensions_data DataFrame contains indices other
            that "minimize", "ideal", or "nadir".
            RequestError: If message is not a str or a list.
            RequestError: If message is a list but one or more elements are not
            str.
        """
        if message is None:
            message = (
                f"Please provide a reference point better than:\n"
                f"{dimensions_data.loc['nadir'].values.tolist()},\n"
                f"but worse than:\n"
                f"{dimensions_data.loc['ideal'].values.tolist()}"
            )
        if preference_validator is None:
            preference_validator = validate_ref_point_with_ideal_and_nadir
        acceptable_dimensions_data_indices = [
            "minimize",  # 1 if minimized, -1 if maximized
            "ideal",
            "nadir",
        ]
        if not isinstance(dimensions_data, (pd.DataFrame, type(None))):
            msg = (
                f"Dimensional data should be in a pandas dataframe.\n"
                f"Provided data is of type: {type(dimensions_data)}"
            )
            raise RequestError(msg)
        rouge_indices = [
            index
            for index in dimensions_data.index
            if index not in acceptable_dimensions_data_indices
        ]
        if rouge_indices:
            msg = (
                f"dimensions_data should only contain the following indices:\n"
                f"{acceptable_dimensions_data_indices}\n"
                f"The dataframe provided contains the following unsupported indices:\n"
                f"{rouge_indices}"
            )
            raise RequestError(msg)
        if not isinstance(message, str):
            if not isinstance(message, list):
                msg = (
                    f"Message/s to be printed should be string or list of strings"
                    f"Message provided is of type: {type(message)}"
                )
                raise RequestError(msg)
            elif not all(isinstance(x, str) for x in message):
                msg = (
                    "Message/s to be printed should be string or list of strings"
                    "Some elements of the list are not strings"
                )
                raise RequestError(msg)
        content = {
            "dimensions_data": dimensions_data,
            "message": message,
            "validator": preference_validator,
        }
        super().__init__(
            request_type="reference_point_preference",
            interaction_priority=interaction_priority,
            content=content,
            request_id=request_id,
        )

    @BaseRequest.response.setter
    def response(self, value: pd.DataFrame):
        """Validate user preference. Accept if it is valid.

        Args:
            value (pd.DataFrame): The user preference in the form of a pandas
            DataFrame

        Raises:
            RequestError: If reference point is not provided in a pandas
            DataFrame.
        """
        if not isinstance(value, pd.DataFrame):
            msg = "Reference should be provided in a pandas dataframe format"
            raise RequestError(msg)
        self.content["validator"](
            reference_point=value, dimensions_data=self.content["dimensions_data"]
        )
        self._response = value


class PreferredSolutionPreference(BaseRequest):
    """
    Methods can use this class to ask the Decision maker to provide their preferences in form of preferred solution(s).
    """

    def __init__(
        self,
        n_solutions: int,
        message: str = None,
        interaction_priority: str = "required",
        preference_validator: Callable = None,
        request_id: int = None,
    ):
        """
        Initialize preference-class with information about problem.

        Args:
            n_solutions (int): Number of solutions in total.
            message (str): Message to be displayed to the Decision maker.
            interaction_priority (str): Level of priority.
            preference_validator (Callable): Function that validates the Decision maker's preferences.
            request_id (int): Identification number of request.
        """

        self._n_solutions = n_solutions

        if message is None:
            message = (
                "Please specify preferred solution(s) by their index as 'preferred_solutions_indices', so that the "
                "indices start at 0. Please specify the index/indices in a list."
            )

        if preference_validator is None:
            preference_validator = validate_specified_solutions

        if not isinstance(message, str):
            if not isinstance(message, list):
                msg = (
                    f"Message/s to be printed should be string or list of strings"
                    f"Message provided is of type: {type(message)}"
                )
                raise RequestError(msg)
            elif not all(isinstance(x, str) for x in message):
                msg = (
                    "Message/s to be printed should be string or list of strings"
                    "Some elements of the list are not strings"
                )
                raise RequestError(msg)
        content = {"message": message, "validator": preference_validator}
        super().__init__(
            request_type="preferred_solution_preference",
            interaction_priority=interaction_priority,
            content=content,
            request_id=request_id,
        )

    @BaseRequest.response.setter
    def response(self, value):
        # validate the response
        self.content["validator"](indices=value, n_solutions=self._n_solutions)
        self._response = value


class NonPreferredSolutionPreference(BaseRequest):
    """
    Methods can use this class to ask the Decision maker to provide their preferences in form of non-preferred
    solution(s).
    """

    def __init__(
        self,
        n_solutions: int,
        message: str = None,
        interaction_priority: str = "required",
        preference_validator: Callable = None,
        request_id: int = None,
    ):
        """
        Initialize preference-class with information about problem.

        Args:
            n_solutions (int): Number of solutions in total.
            message (str): Message to be displayed to the Decision maker.
            interaction_priority (str): Level of priority.
            preference_validator (Callable): Function that validates the Decision maker's preferences.
            request_id (int): Identification number of request.
        """

        self._n_solutions = n_solutions

        if message is None:
            message = (
                "Please specify non-preferred solution(s) by their index as 'non-preferred_solutions_indices', so that "
                "the indices start at 0. Please specify the index/indices in a list."
            )

        if preference_validator is None:
            preference_validator = validate_specified_solutions

        if not isinstance(message, str):
            if not isinstance(message, list):
                msg = (
                    f"Message/s to be printed should be string or list of strings"
                    f"Message provided is of type: {type(message)}"
                )
                raise RequestError(msg)
            elif not all(isinstance(x, str) for x in message):
                msg = (
                    "Message/s to be printed should be string or list of strings"
                    "Some elements of the list are not strings"
                )
                raise RequestError(msg)
        content = {"message": message, "validator": preference_validator}
        super().__init__(
            request_type="non_preferred_solution_preference",
            interaction_priority=interaction_priority,
            content=content,
            request_id=request_id,
        )

    @BaseRequest.response.setter
    def response(self, value):
        # validate the response
        self.content["validator"](indices=value, n_solutions=self._n_solutions)
        self._response = value


class BoundPreference(BaseRequest):
    """
    Methods can use this class to ask the Decision maker to provide their preferences in form of preferred lower and
    upper bounds for objective values.
    """

    def __init__(
        self,
        dimensions_data: pd.DataFrame,
        n_objectives: int,
        message: str = None,
        interaction_priority: str = "required",
        preference_validator: Callable = None,
        request_id: int = None,
    ):
        """
        Initialize preference-class with information about problem.

        Args:
            dimensions_data (pd.DataFrame): DataFrame including information whether an objective is minimized or
            maximized, for each objective. In addition, includes ideal and nadir vectors.
            n_objectives (int): Number of objectives in problem.
            message (str): Message to be displayed to the Decision maker.
            interaction_priority (str): Level of priority.
            preference_validator (Callable): Function that validates the Decision maker's preferences.
            request_id (int): Identification number of request.
        """

        self._n_objectives = n_objectives

        if message is None:
            message = (
                "Please specify desired lower and upper bound for each objective as 'bounds', starting from the first "
                "objective and ending with the last one. Please specify the bounds as a numpy array containing lists, "
                "so that the first item of list is the lower bound and the second the upper bound, for each objective."
                "For example: numpy.array([[1, 2], [2, 5], [0, 3.5]]), for problem with three objectives."
                "Ideal vector: {}\nNadir vector: {}.".format(
                    dimensions_data.loc["ideal"].values.tolist(),
                    dimensions_data.loc["nadir"].values.tolist(),
                )
            )

        if preference_validator is None:
            preference_validator = validate_bounds

        if not isinstance(message, str):
            if not isinstance(message, list):
                msg = (
                    f"Message/s to be printed should be string or list of strings"
                    f"Message provided is of type: {type(message)}"
                )
                raise RequestError(msg)
            elif not all(isinstance(x, str) for x in message):
                msg = (
                    "Message/s to be printed should be string or list of strings"
                    "Some elements of the list are not strings"
                )
                raise RequestError(msg)
        content = {
            "dimensions_data": dimensions_data,
            "message": message,
            "validator": preference_validator,
        }
        super().__init__(
            request_type="bound_preference",
            interaction_priority=interaction_priority,
            content=content,
            request_id=request_id,
        )

    @BaseRequest.response.setter
    def response(self, value):
        # validate the response
        self.content["validator"](
            dimensions_data=self.content["dimensions_data"],
            bounds=value,
            n_objectives=self._n_objectives,
        )
        self._response = value



import numpy as np

from desdeo_tools.scalarization import Scalarizer
from desdeo_tools.solver.ScalarSolver import ScalarMinimizer
from typing import Optional, Callable, Union


class ECMError(Exception):
    """Raised when an error related to the Epsilon Constraint Method is encountered.

    """


class EpsilonConstraintMethod:
    """A class to represent a class for scalarizing MOO problems using the epsilon
        constraint method.
    Attributes:
        objectives (Callable): Objective functions.
        to_be_minimized (int): Integer representing which objective function
        should be minimized.
        epsilons (np.ndarray): Upper bounds chosen by the decison maker.
                               Epsilon constraint functions are defined in a following form:
                                    f_i(x) <= eps_i
                               If the constraint function is of form
                                    f_i(x) >= eps_i
                               Remember to multiply the epsilon value with -1!
        constraints (Optional[Callable]): Function that returns definitions of other constraints, if existing.
    """

    def __init__(
            self, objectives: Callable, to_be_minimized: int, epsilons: np.ndarray,
            constraints: Optional[Callable]
    ):
        self.objectives = objectives
        self._to_be_minimized = to_be_minimized
        self.epsilons = epsilons
        self.constraints = constraints

    def evaluate_constraints(self, xs) -> np.ndarray:
        """
        Returns values of constraints with given decison variables.
        Args:
            xs (np.ndarray): Decision variables.
        Returns: Values of constraint functions (both "original" constraints as well as epsilon constraints) in a vector.
        """
        xs = np.atleast_2d(xs)

        # evaluate epsilon constraint function "left-side" values with given decision variables
        epsilon_left_side = np.array(
            [val for nrow, row in enumerate(self.objectives(xs))
             for ival, val in enumerate(row) if ival != self._to_be_minimized
             ])

        if len(epsilon_left_side) != len(self.epsilons):
            msg = ("The lenght of the epsilons array ({}) must match the total number of objectives - 1 ({})."
                   ).format(len(self.epsilons), len(self.objectives(xs)) - 1)
            raise ECMError(msg)

        # evaluate values of epsilon constraint functions
        e: np.ndarray = np.array([-(f - v) for f, v in zip(epsilon_left_side, self.epsilons)])

        if self.constraints(xs) is not None:
            c = self.constraints(xs)
            return np.concatenate([c, e], axis=None)  # does it work with multiple constraints?
        else:
            return e

    def __call__(self, objective_vector: np.ndarray) -> Union[float, np.ndarray]:
        """
        Returns the value of objective function to be minimized.
        Args:
            objective_vector (np.ndarray): Values of objective functions.

        Returns: Value of objective function to be minimized.
        """
        if np.shape(objective_vector)[0] > 1:  # more rows than one
            return np.array([objective_vector[i][self._to_be_minimized] for i, _ in enumerate(objective_vector)])
        else:
            return objective_vector[0][self._to_be_minimized]


# Testing the method
if __name__ == "__main__":
    # 1. Define objective functions, bounds and constraints

    def volume(r, h):
        return np.pi * r ** 2 * h


    def area(r, h):
        return 2 * np.pi ** 2 + np.pi * r * h


    # add third objective
    def weight(v):
        return 0.01 * v


    def objective(xs):
        # xs is a 2d array like, which has different values for r and h on its first and second columns respectively.
        xs = np.atleast_2d(xs)
        return np.stack((volume(xs[:, 0], xs[:, 1]), -area(xs[:, 0], xs[:, 1]), weight(volume(xs[:, 0], xs[:, 1])))).T


    # bounds for decision variables
    r_bounds = np.array([2.5, 15])
    h_bounds = np.array([10, 50])
    bounds = np.stack((r_bounds, h_bounds))


    # constraints
    def con_golden(xs):
        # constraints are defined in DESDEO in a way were a positive value indicates an agreement with a constraint, and
        # a negative one a disagreement.
        xs = np.atleast_2d(xs)
        return -(xs[:, 0] / xs[:, 1] - 1.618)


    # 2. Apply Epsilon contraint method

    # index of which objective function to minimize
    obj_min = 2

    # set upper bound(s) for the other objectives, in the same order than which corresponding objective functions
    # are defined
    epsil = np.array([2000, -100])  # multiply the epsilons with -1, if the constraint is of form f_i(x) >= e_i

    # create an instance of EpsilonConstraintMethod-class for given problem
    eps = EpsilonConstraintMethod(objective, obj_min, epsil, constraints=con_golden)

    # constraint evaluator, used by the solver
    cons_evaluate = eps.evaluate_constraints

    # scalarize
    scalarized_objective = Scalarizer(objective, eps)
    print(scalarized_objective)

    # 3. Solve

    # starting point
    x0 = np.array([2, 11])
    minimizer = ScalarMinimizer(scalarized_objective, bounds, constraint_evaluator=cons_evaluate, method=None)

    # minimize
    res = minimizer.minimize(x0)
    final_r, final_h = res["x"][0], res["x"][1]
    final_obj = objective(res["x"]).squeeze()
    final_V, final_A, final_W = final_obj[0], final_obj[1], final_obj[2]

    print(f"Final cake specs: radius: {final_r}cm, height: {final_h}cm.")
    print(f"Final cake dimensions: volume: {final_V}, area: {-final_A}, weight: {final_W}.")
    print(final_r / final_h)
    print(res)


"""This module implements methods for defining functions to scalarize vector valued functions.
These are knows as `Scalarizer`s.
It also provides achievement
scalarizing functions to be used with the scalarizers.

"""

__all__ = [
    "AugmentedGuessASF",
    "MaxOfTwoASF",
    "PointMethodASF",
    "ReferencePointASF",
    "SimpleASF",
    "StomASF",
    "DiscreteScalarizer",
    "Scalarizer",
    "Tchebycheff",
    "WeightedSum",
    "PBI",
    "reference_point_method_GLIDE",
    "AUG_GUESS_GLIDE",
    "GUESS_GLIDE",
    "AUG_STOM_GLIDE",
    "STOM_GLIDE",
    "NIMBUS_GLIDE",
    "PROJECT_GLIDE",
    "STEP_GLIDE",
    "Tchebycheff_GLIDE",
]

from desdeo_tools.scalarization.ASF import (
    AugmentedGuessASF,
    MaxOfTwoASF,
    PointMethodASF,
    ReferencePointASF,
    SimpleASF,
    StomASF,
)
from desdeo_tools.scalarization.Scalarizer import DiscreteScalarizer, Scalarizer
from desdeo_tools.scalarization.MOEADSF import Tchebycheff, WeightedSum, PBI

from desdeo_tools.scalarization.GLIDE_II import (
    reference_point_method_GLIDE,
    AUG_GUESS_GLIDE,
    GUESS_GLIDE,
    AUG_STOM_GLIDE,
    STOM_GLIDE,
    NIMBUS_GLIDE,
    PROJECT_GLIDE,
    STEP_GLIDE,
    Tchebycheff_GLIDE,
)
